from __future__ import annotations

import asyncio
import contextlib
import contextvars
import dataclasses
import logging
import sys
import uuid
from collections.abc import Awaitable
from concurrent.futures import ThreadPoolExecutor
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from academy.agent import AgentT
from academy.context import ActionContext
from academy.context import AgentContext
from academy.exception import ActionCancelledError
from academy.exception import ExchangeError
from academy.exception import MailboxTerminatedError
from academy.exception import raise_exceptions
from academy.exchange import AgentExchangeClient
from academy.exchange import ExchangeClient
from academy.exchange import ExchangeFactory
from academy.exchange.transport import AgentRegistrationT
from academy.exchange.transport import ExchangeTransportT
from academy.handle import exchange_context
from academy.identifier import EntityId
from academy.message import ActionRequest
from academy.message import ActionResponse
from academy.message import ErrorResponse
from academy.message import Message
from academy.message import PingRequest
from academy.message import Request
from academy.message import Response
from academy.message import ResponseT_co
from academy.message import ShutdownRequest
from academy.message import SuccessResponse
from academy.serialize import NoPickleMixin

logger = logging.getLogger(__name__)

T = TypeVar('T')


@dataclasses.dataclass
class _ShutdownState:
    # If the shutdown was expected or due to an error
    expected_shutdown: bool = True
    # Override the termination setting of the run config
    terminate_override: bool | None = None


@dataclasses.dataclass(frozen=True)
class RuntimeConfig:
    """Agent runtime configuration.

    Attributes:
        cancel_actions_on_shutdown: Cancel running actions when the agent
            is shutdown, otherwise wait for the actions to finish.
        max_sync_concurrency: Maximum number of concurrent sync tasks allowed
            via [`Agent.agent_run_sync()`][academy.agent.Agent.agent_run_sync].
            This is used to set the number of threads in a default
            [`ThreadPoolExecutor`][concurrent.futures.ThreadPoolExecutor].
        raise_loop_errors_on_shutdown: Raise any captured loop errors when
            the agent is shutdown.
        shutdown_on_loop_error: Shutdown the agent if any loop raises an error.
        terminate_on_error: Terminate the agent by closing its mailbox
            permanently if the agent shuts down due to an error.
        terminate_on_success: Terminate the agent by closing its mailbox
            permanently if the agent shuts down without an error.
    """

    cancel_actions_on_shutdown: bool = True
    max_sync_concurrency: int | None = None
    raise_loop_errors_on_shutdown: bool = True
    shutdown_on_loop_error: bool = True
    terminate_on_error: bool = True
    terminate_on_success: bool = True


class Runtime(Generic[AgentT], NoPickleMixin):
    """Agent runtime manager.

    The runtime is used to execute an agent by managing stateful resources,
    startup/shutdown, lifecycle hooks, and concurrency.

    An agent can be run in two ways:
    ```python
    runtime = Runtime(agent, ...)

    # Option 1: Async context manager
    async with runtime:
        ...
        await runtime.wait_shutdown()

    # Option 2: Run until complete
    await runtime.run_until_complete()
    ```

    Note:
        A runtime can only be used once, after which attempts to run an
        agent using the same runtime with raise a
        [`RuntimeError`][RuntimeError].

    Note:
        If any `@loop` method raises an error, the agent will be signaled
        to shutdown if `shutdown_on_loop_error` is set in the `config`.

    Args:
        agent: Agent that the agent will exhibit.
        exchange_factory: Message exchange factory.
        registration: Agent registration info returned by the exchange.
        config: Agent execution parameters.
    """

    def __init__(
        self,
        agent: AgentT,
        *,
        exchange_factory: ExchangeFactory[ExchangeTransportT],
        registration: AgentRegistrationT,
        config: RuntimeConfig | None = None,
    ) -> None:
        self.agent_id = registration.agent_id
        self.agent = agent
        self.factory = exchange_factory
        self.registration = registration
        self.config = config if config is not None else RuntimeConfig()

        self._actions = agent._agent_actions()
        self._loops = agent._agent_loops()

        self._started_event = asyncio.Event()
        self._shutdown_event = asyncio.Event()
        self._shutdown_options = _ShutdownState()
        self._agent_startup_called = False

        self._action_tasks: dict[uuid.UUID, asyncio.Task[None]] = {}
        self._loop_tasks: dict[str, asyncio.Task[None]] = {}
        self._loop_exceptions: list[tuple[str, Exception]] = []

        self._sync_executor = ThreadPoolExecutor(
            self.config.max_sync_concurrency,
            thread_name_prefix='agent-sync-executor-thread',
        )

        self._exchange_client: (
            AgentExchangeClient[AgentT, ExchangeTransportT] | None
        ) = None
        self._exchange_listener_task: asyncio.Task[None] | None = None
        self.exchange_context_token: (
            contextvars.Token[ExchangeClient[Any]] | None
        ) = None

    async def __aenter__(self) -> Self:
        try:
            await self._start()
        except:
            self.signal_shutdown(expected=False)
            await self._shutdown()
            raise
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        await self._shutdown()

    def __repr__(self) -> str:
        name = type(self).__name__
        return f'{name}({self.agent!r}, {self._exchange_client!r})'

    def __str__(self) -> str:
        name = type(self).__name__
        agent = type(self.agent).__name__
        return f'{name}<{agent}; {self.agent_id}>'

    async def _send_response(self, response: Message[ResponseT_co]) -> None:
        assert self._exchange_client is not None
        try:
            await self._exchange_client.send(response)
        except MailboxTerminatedError:  # pragma: no cover
            logger.warning(
                'Failed to send response from %s to %s because the '
                'destination mailbox was terminated.',
                self.agent_id,
                response.dest,
            )
        except ExchangeError:  # pragma: no cover
            logger.exception(
                'Failed to send response from %s to %s.',
                self.agent_id,
                response.dest,
            )

    async def _execute_action(self, request: Message[ActionRequest]) -> None:
        body = request.get_body()
        response: Message[Response]
        try:
            # Do not run the method until the startup sequence has finished
            await self._started_event.wait()
            result = await self.action(
                body.action,
                request.src,
                args=body.get_args(),
                kwargs=body.get_kwargs(),
            )
        except asyncio.CancelledError:
            response = request.create_response(
                ErrorResponse(exception=ActionCancelledError(body.action)),
            )
        except Exception as e:
            response = request.create_response(ErrorResponse(exception=e))
        else:
            response = request.create_response(
                ActionResponse(action=body.action, result=result),
            )
        finally:
            # Shield sending the result from being cancelled so the requester
            # does not block on a response they will never get.
            await asyncio.shield(self._send_response(response))

    async def _execute_loop(
        self,
        name: str,
        method: Callable[[asyncio.Event], Awaitable[None]],
    ) -> None:
        try:
            # Do not run the method until the startup sequence has finished
            await self._started_event.wait()
            await method(self._shutdown_event)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self._loop_exceptions.append((name, e))
            logger.exception(
                'Error in loop %r (signaling shutdown: %s)',
                name,
                self.config.shutdown_on_loop_error,
            )
            if self.config.shutdown_on_loop_error:
                self.signal_shutdown(expected=False)

    async def _request_handler(self, request: Message[Request]) -> None:
        body = request.get_body()
        if isinstance(body, ActionRequest):
            task = asyncio.create_task(
                self._execute_action(request),  # type: ignore[arg-type]
                name=f'execute-action-{body.action}-{request.tag}',
            )
            self._action_tasks[request.tag] = task
            task.add_done_callback(
                lambda _: self._action_tasks.pop(request.tag),
            )
        elif isinstance(body, PingRequest):
            logger.info('Ping request received by %s', self.agent_id)
            response = request.create_response(SuccessResponse())
            await self._send_response(response)
        elif isinstance(body, ShutdownRequest):
            self.signal_shutdown(expected=True, terminate=body.terminate)
        else:
            raise AssertionError('Unreachable.')

    async def action(
        self,
        action: str,
        source_id: EntityId,
        *,
        args: Any,
        kwargs: Any,
    ) -> Any:
        """Invoke an action of the agent's agent.

        Args:
            action: Name of action to invoke.
            source_id: ID of the source that requested the action.
            args: Tuple of positional arguments.
            kwargs: Dictionary of keyword arguments.

        Returns:
            Result of the action.

        Raises:
            AttributeError: If an action with this name is not implemented by
                the agent's agent.
        """
        logger.debug('Invoking "%s" action on %s', action, self.agent_id)
        if action not in self._actions:
            raise AttributeError(
                f'{self.agent} does not have an action named "{action}".',
            )
        action_method = self._actions[action]
        if action_method._action_method_context:
            assert self._exchange_client is not None
            context = ActionContext(source_id, self._exchange_client)
            return await action_method(*args, context=context, **kwargs)
        else:
            return await action_method(*args, **kwargs)

    async def run_until_complete(self) -> None:
        """Run the agent until shutdown.

        Agent startup involves:

        1. Creates a new exchange client for the agent, and sets the exchange
           context to this agent's client for all handles.
        1. Sets the runtime context on the agent.
        1. Binds all handles of the agent to this agent's exchange client.
        1. Schedules a [`Task`][asyncio.Task] to listen for messages in the
           agent's mailbox in the exchange. Agent requests will not start
           processing until the end of the startup sequence.
        1. Schedules a [`Task`][asyncio.Task] for all control loops defined on
           the agent. Each task will block until the end of the startup
           sequence before starting the loop.
        1. Calls
           [`Agent.agent_on_startup()`][academy.agent.Agent.agent_on_startup].

        After startup succeeds, this method waits for the agent to be shutdown,
        such as due to a failure in a control loop or receiving a shutdown
        message.

        Agent shutdown involves:

        1. Calls
           [`Agent.agent_on_shutdown()`][academy.agent.Agent.agent_on_shutdown].
        1. Cancels running control loop tasks.
        1. Cancels the mailbox message listener task so no new requests are
           received.
        1. Waits for any currently executing actions to complete.
        1. Terminates the agent's mailbox in the exchange if configured.
        1. Restores the exchange context and closes the exchange client.

        Raises:
            RuntimeError: If the agent has already been shutdown.
            Exception: Any exceptions raised during startup, shutdown, or
                inside of control loops.
        """
        async with self:
            await self.wait_shutdown()

    async def _start(self) -> None:
        if self._shutdown_event.is_set():
            raise RuntimeError('Agent has already been shutdown.')

        logger.debug(
            'Starting agent... (%s; %s)',
            self.agent_id,
            self.agent,
        )

        self._exchange_client = await self.factory.create_agent_client(
            self.registration,
            request_handler=self._request_handler,
        )
        self.exchange_context_token = exchange_context.set(
            self._exchange_client,
        )

        context = AgentContext(
            agent_id=self.agent_id,
            exchange_client=self._exchange_client,
            executor=self._sync_executor,
            shutdown_event=self._shutdown_event,
        )
        self.agent._agent_set_context(context)

        self._exchange_listener_task = asyncio.create_task(
            self._exchange_client._listen_for_messages(),
            name=f'exchange-listener-{self.agent_id}',
        )

        for name, method in self._loops.items():
            task = asyncio.create_task(
                self._execute_loop(name, method),
                name=f'execute-loop-{name}-{self.agent_id}',
            )
            self._loop_tasks[name] = task

        await self.agent.agent_on_startup()
        self._agent_startup_called = True

        self._started_event.set()
        logger.info('Running agent (%s; %s)', self.agent_id, self.agent)

    def _should_terminate_mailbox(self) -> bool:
        # Inspects the shutdown options and the run config to determine
        # if the agent's mailbox should be terminated in the exchange.
        if self._shutdown_options.terminate_override is not None:
            return self._shutdown_options.terminate_override

        expected = self._shutdown_options.expected_shutdown
        terminate_for_success = self.config.terminate_on_success and expected
        terminate_for_error = self.config.terminate_on_error and not expected
        return terminate_for_success or terminate_for_error

    async def _shutdown(self) -> None:
        logger.debug(
            'Shutting down agent... (expected: %s; %s; %s)',
            self._shutdown_options.expected_shutdown,
            self.agent_id,
            self.agent,
        )

        self._shutdown_event.set()

        if self._agent_startup_called:
            # Don't call agent_on_shutdown() if we never called
            # agent_on_startup()
            await self.agent.agent_on_shutdown()

        # Cancel running control loop tasks
        for task in self._loop_tasks.values():
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

        # If _start() fails early, the listener task may not have started.
        if self._exchange_listener_task is not None:
            # Stop exchange listener thread before cancelling waiting on
            # running actions so we know that we won't receive an new
            # action requests
            self._exchange_listener_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._exchange_listener_task

        # Wait for running actions to complete
        for task in tuple(self._action_tasks.values()):
            # Both branches should be covered by
            # test_agent_action_message_cancelled but a slow test runner could
            # not begin shutdown until all the tasks have completed anyways
            if self.config.cancel_actions_on_shutdown:  # pragma: no branch
                task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

        self._sync_executor.shutdown()
        # Reset exchange client (for the case of nested execution)
        if self.exchange_context_token is not None:
            exchange_context.reset(self.exchange_context_token)
            self.exchange_context_token = None

        if self._exchange_client is not None:
            if self._should_terminate_mailbox():
                await self._exchange_client.terminate(self.agent_id)
            await self._exchange_client.close()

        if self.config.raise_loop_errors_on_shutdown:
            # Raise loop exceptions so the caller sees them, even if the loop
            # errors didn't cause the shutdown.
            raise_exceptions(
                (e for _, e in self._loop_exceptions),
                message='Caught failures in agent loops while shutting down.',
            )

        logger.info('Shutdown agent (%s; %s)', self.agent_id, self.agent)

    def signal_shutdown(
        self,
        *,
        expected: bool = True,
        terminate: bool | None = None,
    ) -> None:
        """Signal that the agent should exit.

        If the agent has not started, this will cause the agent to immediately
        shutdown when next started. If the agent is shutdown, this has no
        effect.

        Args:
            expected: If the reason for the shutdown was due to normal
                expected reasons or due to unexpected errors.
            terminate: Optionally override the mailbox termination settings
                in the run config.
        """
        self._shutdown_options = _ShutdownState(
            expected_shutdown=expected,
            terminate_override=terminate,
        )
        self._shutdown_event.set()

    async def wait_shutdown(self, timeout: float | None = None) -> None:
        """Wait for agent shutdown to be signalled.

        Args:
            timeout: Optional timeout in seconds to wait for a shutdown event.

        Raises:
            TimeoutError: If `timeout` was exceeded while waiting for agents.
        """
        try:
            await asyncio.wait_for(
                self._shutdown_event.wait(),
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            # In Python 3.10 and older, asyncio.TimeoutError and TimeoutError
            # are different error types.
            raise TimeoutError(
                f'Agent shutdown was not signalled within {timeout} seconds.',
            ) from None



from __future__ import annotations

import asyncio
from typing import TypeVar

from academy.agent import action
from academy.agent import Agent
from academy.agent import loop

T = TypeVar('T')


class EmptyAgent(Agent):
    pass


class ErrorAgent(Agent):
    @action
    async def fails(self) -> None:
        raise RuntimeError('This action always fails.')


class IdentityAgent(Agent):
    @action
    async def identity(self, value: T) -> T:
        return value


class WaitAgent(Agent):
    @loop
    async def wait(self, shutdown: asyncio.Event) -> None:
        await shutdown.wait()


class CounterAgent(Agent):
    def __init__(self) -> None:
        super().__init__()
        self._count = 0

    async def agent_on_startup(self) -> None:
        self._count = 0

    @action
    async def add(self, value: int) -> None:
        self._count += value

    @action
    async def count(self) -> int:
        return self._count


class SleepAgent(Agent):
    def __init__(self, loop_sleep: float = 0.001) -> None:
        super().__init__()
        self.loop_sleep = loop_sleep
        self.steps = 0

    @loop
    async def count(self, shutdown: asyncio.Event) -> None:
        while not shutdown.is_set():
            self.steps += 1
            await asyncio.sleep(self.loop_sleep)



from __future__ import annotations

import logging
import multiprocessing
import pathlib
import time
import uuid
from collections.abc import AsyncGenerator
from typing import Any
from unittest import mock

import pytest
import pytest_asyncio
from aiohttp.test_utils import TestClient
from aiohttp.test_utils import TestServer
from aiohttp.web import Application
from aiohttp.web import Request

from academy.exception import BadEntityIdError
from academy.exception import MailboxTerminatedError
from academy.exchange import HttpExchangeFactory
from academy.exchange import MailboxStatus
from academy.exchange.cloud.config import ExchangeAuthConfig
from academy.exchange.cloud.config import ExchangeServingConfig
from academy.exchange.cloud.exceptions import ForbiddenError
from academy.exchange.cloud.login import AcademyExchangeScopes
from academy.exchange.cloud.server import _MailboxManager
from academy.exchange.cloud.server import _main
from academy.exchange.cloud.server import _run
from academy.exchange.cloud.server import create_app
from academy.exchange.cloud.server import StatusCode
from academy.identifier import AgentId
from academy.identifier import UserId
from academy.message import Message
from academy.message import PingRequest
from academy.socket import open_port
from testing.constant import TEST_SLEEP_INTERVAL
from testing.ssl import SSLContextFixture


def test_server_cli(tmp_path: pathlib.Path) -> None:
    data = """\
host = "localhost"
port = 1234
certfile = "/path/to/cert.pem"
keyfile = "/path/to/privkey.pem"

[auth]
method = "globus"

[auth.kwargs]
client_id = "ABC"
"""

    filepath = tmp_path / 'exchange.toml'
    with open(filepath, 'w') as f:
        f.write(data)

    with mock.patch('academy.exchange.cloud.server._run'):
        assert _main(['--config', str(filepath)]) == 0


def test_server_run() -> None:
    config = ExchangeServingConfig(
        host='127.0.0.1',
        port=open_port(),
        log_level=logging.ERROR,
    )
    context = multiprocessing.get_context('spawn')
    process = context.Process(target=_run, args=(config,))
    process.start()

    while True:
        try:
            client = HttpExchangeFactory(
                f'http://{config.host}:{config.port}/',
            ).create_user_client()
        except OSError:  # pragma: no cover
            time.sleep(TEST_SLEEP_INTERVAL)
        else:
            client.close()
            break

    process.terminate()
    process.join()


@pytest.mark.filterwarnings('ignore:Unverified HTTPS request is being made')
def test_server_run_ssl(ssl_context: SSLContextFixture) -> None:
    config = ExchangeServingConfig(
        host='127.0.0.1',
        port=open_port(),
        log_level=logging.ERROR,
    )
    config.certfile = ssl_context.certfile
    config.keyfile = ssl_context.keyfile

    context = multiprocessing.get_context('spawn')
    process = context.Process(target=_run, args=(config,))
    process.start()

    while True:
        try:
            client = HttpExchangeFactory(
                f'https://{config.host}:{config.port}/',
                ssl_verify=False,
            ).create_user_client()
        except OSError:  # pragma: no cover
            time.sleep(TEST_SLEEP_INTERVAL)
        else:
            client.close()
            break

    process.terminate()
    process.join()


@pytest.mark.asyncio
async def test_mailbox_manager_create_close() -> None:
    manager = _MailboxManager()
    user_id = str(uuid.uuid4())
    uid = UserId.new()
    # Should do nothing since mailbox doesn't exist
    await manager.terminate(user_id, uid)
    assert await manager.check_mailbox(user_id, uid) == MailboxStatus.MISSING
    manager.create_mailbox(user_id, uid)
    assert await manager.check_mailbox(user_id, uid) == MailboxStatus.ACTIVE
    manager.create_mailbox(user_id, uid)  # Idempotent check

    bad_user = str(uuid.uuid4())  # Authentication check
    with pytest.raises(ForbiddenError):
        manager.create_mailbox(bad_user, uid)
    with pytest.raises(ForbiddenError):
        await manager.check_mailbox(bad_user, uid)
    with pytest.raises(ForbiddenError):
        await manager.terminate(bad_user, uid)

    await manager.terminate(user_id, uid)
    await manager.terminate(user_id, uid)  # Idempotent check


@pytest.mark.asyncio
async def test_mailbox_manager_send_recv() -> None:
    manager = _MailboxManager()
    user_id = str(uuid.uuid4())
    bad_user = str(uuid.uuid4())
    uid = UserId.new()
    manager.create_mailbox(user_id, uid)

    message = Message.create(src=uid, dest=uid, body=PingRequest())
    with pytest.raises(ForbiddenError):
        await manager.put(bad_user, message)
    await manager.put(user_id, message)

    with pytest.raises(ForbiddenError):
        await manager.get(bad_user, uid)
    assert await manager.get(user_id, uid) == message

    await manager.terminate(user_id, uid)


@pytest.mark.asyncio
async def test_mailbox_manager_bad_identifier() -> None:
    manager = _MailboxManager()
    uid = UserId.new()
    message = Message.create(src=uid, dest=uid, body=PingRequest())

    with pytest.raises(BadEntityIdError):
        await manager.get(None, uid)

    with pytest.raises(BadEntityIdError):
        await manager.put(None, message)


@pytest.mark.asyncio
async def test_mailbox_manager_mailbox_closed() -> None:
    manager = _MailboxManager()
    uid = UserId.new()
    manager.create_mailbox(None, uid)
    await manager.terminate(None, uid)
    message = Message.create(src=uid, dest=uid, body=PingRequest())

    with pytest.raises(MailboxTerminatedError):
        await manager.get(None, uid)

    with pytest.raises(MailboxTerminatedError):
        await manager.put(None, message)


@pytest_asyncio.fixture
async def cli() -> AsyncGenerator[TestClient[Request, Application]]:
    app = create_app()
    async with TestClient(TestServer(app)) as client:
        yield client


@pytest.mark.asyncio
async def test_create_mailbox_validation_error(cli) -> None:
    response = await cli.post('/mailbox', json={'mailbox': 'foo'})
    assert response.status == StatusCode.BAD_REQUEST.value
    assert await response.text() == 'Missing or invalid mailbox ID'


@pytest.mark.asyncio
async def test_terminate_validation_error(cli) -> None:
    response = await cli.delete('/mailbox', json={'mailbox': 'foo'})
    assert response.status == StatusCode.BAD_REQUEST.value
    assert await response.text() == 'Missing or invalid mailbox ID'


@pytest.mark.asyncio
async def test_discover_validation_error(cli) -> None:
    response = await cli.get('/discover', json={})
    assert response.status == StatusCode.BAD_REQUEST.value
    assert await response.text() == 'Missing or invalid arguments'


@pytest.mark.asyncio
async def test_check_mailbox_validation_error(cli) -> None:
    response = await cli.get('/mailbox', json={'mailbox': 'foo'})
    assert response.status == StatusCode.BAD_REQUEST.value
    assert await response.text() == 'Missing or invalid mailbox ID'


@pytest.mark.asyncio
async def test_send_mailbox_validation_error(cli) -> None:
    response = await cli.put('/message', json={'message': 'foo'})
    assert response.status == StatusCode.BAD_REQUEST.value
    assert await response.text() == 'Missing or invalid message'


@pytest.mark.asyncio
async def test_recv_mailbox_validation_error(cli) -> None:
    response = await cli.get('/message', json={'mailbox': 'foo'})
    assert response.status == StatusCode.BAD_REQUEST.value
    assert await response.text() == 'Missing or invalid mailbox ID'

    response = await cli.get(
        '/message',
        json={'mailbox': UserId.new().model_dump_json()},
    )
    assert response.status == StatusCode.NOT_FOUND.value
    assert await response.text() == 'Unknown mailbox ID'


@pytest.mark.asyncio
async def test_recv_timeout_error(cli) -> None:
    uid = UserId.new()
    response = await cli.post(
        '/mailbox',
        json={'mailbox': uid.model_dump_json()},
    )
    assert response.status == StatusCode.OKAY.value

    response = await cli.get(
        '/message',
        json={
            'mailbox': uid.model_dump_json(),
            'timeout': TEST_SLEEP_INTERVAL,
        },
    )
    assert response.status == StatusCode.TIMEOUT.value


@pytest.mark.asyncio
async def test_null_auth_client() -> None:
    auth = ExchangeAuthConfig()
    app = create_app(auth)
    async with TestClient(TestServer(app)) as client:
        response = await client.get('/message', json={'mailbox': 'foo'})
        assert response.status == StatusCode.BAD_REQUEST.value
        assert await response.text() == 'Missing or invalid mailbox ID'

        response = await client.get(
            '/message',
            json={'mailbox': UserId.new().model_dump_json()},
        )
        assert response.status == StatusCode.NOT_FOUND.value
        assert await response.text() == 'Unknown mailbox ID'


@pytest_asyncio.fixture
async def auth_client() -> AsyncGenerator[TestClient[Request, Application]]:
    auth = ExchangeAuthConfig(
        method='globus',
        kwargs={'client_id': str(uuid.uuid4()), 'client_secret': ''},
    )
    user_1: dict[str, Any] = {
        'active': True,
        'username': 'username',
        'client_id': str(uuid.uuid4()),
        'email': 'username@example.com',
        'name': 'User Name',
        'aud': [AcademyExchangeScopes.resource_server],
    }

    user_2: dict[str, Any] = {
        'active': True,
        'username': 'username',
        'client_id': str(uuid.uuid4()),
        'email': 'username@example.com',
        'name': 'User Name',
        'aud': [AcademyExchangeScopes.resource_server],
    }

    inactive: dict[str, Any] = {
        'active': False,
    }

    def authorize(token):
        if token == 'user_1':
            return user_1
        if token == 'user_2':
            return user_2
        else:
            return inactive

    with mock.patch(
        'globus_sdk.ConfidentialAppAuthClient.oauth2_token_introspect',
    ) as mock_token_response:
        mock_token_response.side_effect = authorize
        app = create_app(auth)
        async with TestClient(TestServer(app)) as client:
            yield client


@pytest.mark.asyncio
async def test_globus_auth_client_create_discover_close(auth_client) -> None:
    aid = AgentId.new(name='test').model_dump_json()

    # Create agent
    response = await auth_client.post(
        '/mailbox',
        json={'mailbox': aid, 'agent': 'foo'},
        headers={'Authorization': 'Bearer user_1'},
    )
    assert response.status == StatusCode.OKAY.value

    response = await auth_client.post(
        '/mailbox',
        json={'mailbox': aid, 'agent': 'foo'},
        headers={'Authorization': 'Bearer user_2'},
    )
    assert response.status == StatusCode.FORBIDDEN.value

    # Discover
    response = await auth_client.get(
        '/discover',
        json={'agent': 'foo', 'allow_subclasses': False},
        headers={'Authorization': 'Bearer user_1'},
    )
    response_json = await response.json()
    agent_ids = [
        aid for aid in response_json['agent_ids'].split(',') if len(aid) > 0
    ]
    assert len(agent_ids) == 1
    assert response.status == StatusCode.OKAY.value

    response = await auth_client.get(
        '/discover',
        json={'agent': 'foo', 'allow_subclasses': False},
        headers={'Authorization': 'Bearer user_2'},
    )
    response_json = await response.json()
    agent_ids = [
        aid for aid in response_json['agent_ids'].split(',') if len(aid) > 0
    ]
    assert len(agent_ids) == 0
    assert response.status == StatusCode.OKAY.value

    # Check mailbox
    response = await auth_client.get(
        '/mailbox',
        json={'mailbox': aid},
        headers={'Authorization': 'Bearer user_1'},
    )
    assert response.status == StatusCode.OKAY.value

    response = await auth_client.get(
        '/mailbox',
        json={'mailbox': aid},
        headers={'Authorization': 'Bearer user_2'},
    )
    assert response.status == StatusCode.FORBIDDEN.value

    # Delete mailbox
    response = await auth_client.delete(
        '/mailbox',
        json={'mailbox': aid},
        headers={'Authorization': 'Bearer user_2'},
    )
    assert response.status == StatusCode.FORBIDDEN.value

    response = await auth_client.delete(
        '/mailbox',
        json={'mailbox': aid},
        headers={'Authorization': 'Bearer user_1'},
    )
    assert response.status == StatusCode.OKAY.value


@pytest.mark.asyncio
async def test_globus_auth_client_message(auth_client) -> None:
    aid: AgentId[Any] = AgentId.new(name='test')
    cid = UserId.new()
    message = Message.create(src=cid, dest=aid, body=PingRequest())

    # Create agent
    response = await auth_client.post(
        '/mailbox',
        json={'mailbox': aid.model_dump_json(), 'agent': 'foo'},
        headers={'Authorization': 'Bearer user_1'},
    )
    assert response.status == StatusCode.OKAY.value

    # Create client
    response = await auth_client.post(
        '/mailbox',
        json={'mailbox': cid.model_dump_json()},
        headers={'Authorization': 'Bearer user_1'},
    )
    assert response.status == StatusCode.OKAY.value

    # Send valid message
    response = await auth_client.put(
        '/message',
        json={'message': message.model_dump_json()},
        headers={'Authorization': 'Bearer user_1'},
    )
    assert response.status == StatusCode.OKAY.value

    # Send unauthorized message
    response = await auth_client.put(
        '/message',
        json={'message': message.model_dump_json()},
        headers={'Authorization': 'Bearer user_2'},
    )
    assert response.status == StatusCode.FORBIDDEN.value

    response = await auth_client.get(
        '/message',
        json={'mailbox': aid.model_dump_json()},
        headers={'Authorization': 'Bearer user_1'},
    )
    assert response.status == StatusCode.OKAY.value

    response = await auth_client.get(
        '/message',
        json={'mailbox': aid.model_dump_json()},
        headers={'Authorization': 'Bearer user_2'},
    )
    assert response.status == StatusCode.FORBIDDEN.value


@pytest.mark.asyncio
async def test_globus_auth_client_missing_auth(auth_client) -> None:
    response = await auth_client.get(
        '/discover',
        json={'agent': 'foo', 'allow_subclasses': False},
    )
    assert response.status == StatusCode.UNAUTHORIZED.value


@pytest.mark.asyncio
async def test_globus_auth_client_forbidden(auth_client) -> None:
    response = await auth_client.get(
        '/discover',
        json={'agent': 'foo', 'allow_subclasses': False},
        headers={'Authorization': 'Bearer user_3'},
    )
    assert response.status == StatusCode.FORBIDDEN.value



from __future__ import annotations

import pathlib
import shelve
from typing import Literal
from typing import TypeVar

DEFAULT_PICKLE_PROTOCOL = 5
ValueT = TypeVar('ValueT')


class FileState(shelve.DbfilenameShelf[ValueT]):
    """Dictionary interface for persistent state.

    Persists arbitrary Python objects to disk using [pickle][pickle] and
    a [dbm][dbm] database.

    Note:
        This class uses the [shelve][shelve] module so refer there for
        additional caveats.

    Example:
        ```python
        from typing import Any
        from academy.agent import Agent, action
        from academy.state import FileState

        class Example(Agent):
            def __init__(self, state_path: str) -> None:
                super().__init__()
                self.state: FileState[Any] = FileState(state_path)

            def agent_on_shutdown(self) -> None:
                self.state.close()

            @action
            def get_state(self, key: str) -> Any:
                return self.state[key]

            @action
            def modify_state(self, key: str, value: Any) -> None:
                self.state[key] = value
        ```

    Note:
        When using the [`Manager`][academy.manager.Manager], use the
        deferred agent initialization to ensure that the state is
        initialized on the worker, rather than the client.

        ```python
        from academy.manager import Manager

        async with Manager(...) as manager:
            handle = manager.launch(Example, args=('/tmp/agent-state.dbm',))
        ```

    Args:
        filename: Base filename for the underlying databased used to store
            key-value pairs.
        flag: Open an existing database read-only: `r`; open an existing
            database for read and write: `w`; open a database for read and
            write, creating it if not existent: `c` (default); always create
            a new empty database for read and write: `n`.
        protocol: Pickling protocol. Defaults to version 5; `None` uses
            the [pickle][pickle] default version.
        writeback: By default (`False`), modified objects are only written
            when assigned. If `True`, the object will hold a cache of all
            entries accessed and write them back to the dict at sync and close
            times. This allows natural operations on mutable entries, but can
            consume much more memory and make sync and close take a long time.
    """

    def __init__(
        self,
        filename: str | pathlib.Path,
        *,
        flag: Literal['r', 'w', 'c', 'n'] = 'c',
        protocol: int | None = DEFAULT_PICKLE_PROTOCOL,
        writeback: bool = False,
    ) -> None:
        super().__init__(
            str(filename),
            flag=flag,
            protocol=protocol,
            writeback=writeback,
        )



from __future__ import annotations

import uuid
from typing import Any

from academy.identifier import AgentId
from academy.identifier import UserId


def tests_identifier_equality() -> None:
    aid: AgentId[Any] = AgentId.new()
    cid = UserId.new()

    assert isinstance(aid, AgentId)
    assert isinstance(cid, UserId)

    assert aid != cid
    assert hash(aid) != hash(cid)

    uid = uuid.uuid4()
    aid1: AgentId[Any] = AgentId(uid=uid)
    aid2: AgentId[Any] = AgentId(uid=uid)
    cid1 = UserId(uid=uid)

    assert aid1 == aid2
    assert str(aid1) == str(aid2)
    assert aid1 != cid1


def tests_identifier_equality_ignore_name() -> None:
    uid = uuid.uuid4()
    aid1: AgentId[Any] = AgentId(uid=uid, name='aid1')
    aid2: AgentId[Any] = AgentId(uid=uid, name='aid2')

    assert aid1 == aid2
    assert str(aid1) != str(aid2)



from __future__ import annotations

import base64
import pickle
import sys
import uuid
from typing import Any
from typing import Generic
from typing import get_args
from typing import Literal
from typing import Optional
from typing import TypeVar
from typing import Union

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_serializer
from pydantic import SkipValidation
from pydantic import TypeAdapter

from academy.identifier import EntityId

DEFAULT_FROZEN_CONFIG = ConfigDict(
    arbitrary_types_allowed=True,
    extra='forbid',
    frozen=True,
    use_enum_values=True,
    validate_default=True,
)
DEFAULT_MUTABLE_CONFIG = ConfigDict(
    arbitrary_types_allowed=True,
    extra='forbid',
    frozen=False,
    use_enum_values=True,
    validate_default=True,
)


class ActionRequest(BaseModel):
    """Agent action request message.

    Warning:
        The positional and keywords arguments for the invoked action are
        pickled and base64-encoded when serialized to JSON. This can have
        non-trivial time and space overheads for large arguments.
    """

    action: str = Field(description='Name of the requested action.')
    pargs: SkipValidation[tuple[Any, ...]] = Field(
        default_factory=tuple,
        description='Positional arguments to the action method.',
    )
    kargs: SkipValidation[dict[str, Any]] = Field(
        default_factory=dict,
        description='Keyword arguments to the action method.',
    )
    kind: Literal['action-request'] = Field('action-request', repr=False)

    model_config = DEFAULT_MUTABLE_CONFIG

    @field_serializer('pargs', 'kargs', when_used='json')
    def _pickle_and_encode_obj(self, obj: Any) -> str:
        if isinstance(obj, str):  # pragma: no cover
            return obj
        raw = pickle.dumps(obj)
        return base64.b64encode(raw).decode('utf-8')

    def get_args(self) -> tuple[Any, ...]:
        """Get the positional arguments.

        Lazily deserializes and returns the positional arguments.
        Caches the result to avoid redundant decoding.

        Returns:
            The deserialized tuple of positional arguments.
        """
        if isinstance(self.pargs, str):
            self.pargs = pickle.loads(base64.b64decode(self.pargs))
        return self.pargs

    def get_kwargs(self) -> dict[str, Any]:
        """Get the keyword arguments.

        Lazily deserializes and returns the keyword arguments.
        Caches the result to avoid redundant decoding.

        Returns:
            The deserialized dictionary of keyword arguments.
        """
        if isinstance(self.kargs, str):
            self.kargs = pickle.loads(base64.b64decode(self.kargs))
        return self.kargs


class PingRequest(BaseModel):
    """Agent ping request message."""

    kind: Literal['ping-request'] = Field('ping-request', repr=False)

    model_config = DEFAULT_FROZEN_CONFIG


class ShutdownRequest(BaseModel):
    """Agent shutdown request message."""

    terminate: Optional[bool] = Field(  # noqa: UP045
        None,
        description='Override the termination behavior of the agent.',
    )
    kind: Literal['shutdown-request'] = Field('shutdown-request', repr=False)

    model_config = DEFAULT_FROZEN_CONFIG


class ActionResponse(BaseModel):
    """Agent action response message.

    Warning:
        The result is pickled and base64-encoded when serialized to JSON.
        This can have non-trivial time and space overheads for large results.
    """

    action: str = Field(description='Name of the requested action.')
    result: SkipValidation[Any] = Field(
        description='Result of the action, if successful.',
    )
    kind: Literal['action-response'] = Field('action-response', repr=False)

    model_config = DEFAULT_MUTABLE_CONFIG

    @field_serializer('result', when_used='json')
    def _pickle_and_encode_result(self, obj: Any) -> Optional[list[Any]]:  # noqa: UP045
        if (
            isinstance(obj, list)
            and len(obj) == 2  # noqa PLR2004
            and obj[0] == '__pickled__'
        ):  # pragma: no cover
            # Prevent double serialization
            return obj

        raw = pickle.dumps(obj)
        # This sential value at the start of the tuple is so we can
        # disambiguate a result that is a str versus the string of a
        # serialized result.
        return ['__pickled__', base64.b64encode(raw).decode('utf-8')]

    def get_result(self) -> Any:
        """Get the result.

        Lazily deserializes and returns the result of the action.
        Caches the result to avoid redundant decoding.

        Returns:
            The deserialized result of the action.
        """
        if (
            isinstance(self.result, list)
            and len(self.result) == 2  # noqa PLR2004
            and self.result[0] == '__pickled__'
        ):
            self.result = pickle.loads(base64.b64decode(self.result[1]))
        return self.result


class ErrorResponse(BaseModel):
    """Error response message.

    Contains the exception raised by a failed request.
    """

    exception: SkipValidation[Exception] = Field(
        description='Exception of the failed request.',
    )
    kind: Literal['error-response'] = Field('error-response', repr=False)

    model_config = DEFAULT_MUTABLE_CONFIG

    @field_serializer('exception', when_used='json')
    def _pickle_and_encode_obj(self, obj: Any) -> Optional[str]:  # noqa: UP045
        raw = pickle.dumps(obj)
        return base64.b64encode(raw).decode('utf-8')

    def get_exception(self) -> Exception:
        """Get the exception.

        Lazily deserializes and returns the exception object.
        Caches the result to avoid redundant decoding.

        Returns:
            The deserialized exception.
        """
        if isinstance(self.exception, str):
            self.exception = pickle.loads(base64.b64decode(self.exception))
        return self.exception


class SuccessResponse(BaseModel):
    """Success response message."""

    kind: Literal['success-response'] = Field('success-response', repr=False)

    model_config = DEFAULT_FROZEN_CONFIG


Request = Union[ActionRequest, PingRequest, ShutdownRequest]
Response = Union[ActionResponse, ErrorResponse, SuccessResponse]
Body = Union[Request, Response]

BodyT = TypeVar('BodyT', bound=Body)
RequestT = TypeVar('RequestT', bound=Request)
RequestT_co = TypeVar('RequestT_co', bound=Request, covariant=True)
ResponseT = TypeVar('ResponseT', bound=Response)
ResponseT_co = TypeVar('ResponseT_co', bound=Response, covariant=True)


class Header(BaseModel):
    """Message metadata header.

    Contains information about the sender, receiver, and message context.
    """

    src: EntityId = Field(description='Message source ID.')
    dest: EntityId = Field(description='Message destination ID.')
    tag: uuid.UUID = Field(
        default_factory=uuid.uuid4,
        description='Unique message tag used to match requests and responses.',
    )
    label: Optional[uuid.UUID] = Field(  # noqa: UP045
        None,
        description=(
            'Optional label used to disambiguate response messages when '
            'multiple objects (i.e., handles) share the same mailbox. '
            'This is a different usage from the `tag`.'
        ),
    )
    kind: Literal['request', 'response']

    model_config = DEFAULT_FROZEN_CONFIG

    def create_response_header(self) -> Self:
        """Create a response header based on the current request header.

        Swaps the source and destination, retains the tag and label,
        and sets the kind to 'response'.

        Returns:
            A new header instance with reversed roles.

        Raises:
            ValueError: If the current header is already a response.
        """
        if self.kind == 'response':
            raise ValueError(
                'Cannot create response header from another response header',
            )
        return type(self)(
            tag=self.tag,
            src=self.dest,
            dest=self.src,
            label=self.label,
            kind='response',
        )


class Message(BaseModel, Generic[BodyT]):
    """A complete message with header and body.

    Wraps a header and a typed request/response body. Supports lazy
    deserialization of message bodies, metadata access, and convenient
    construction.

    Note:
        The body value is ignored when testing equality or hashing an instance
        because the body value could be in either a serialized or
        deserialized state until
        [`get_body()`][academy.message.Message.get_body] is called.
    """

    header: Header
    body: SkipValidation[BodyT] = Field(discriminator='kind')

    model_config = DEFAULT_MUTABLE_CONFIG

    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, Message):
            return False
        # body can be in either serialized or unserialized state so
        # we ignore it from equality comparisons
        return self.header == other.header

    def __hash__(self) -> int:
        return hash((type(self), self.header))

    @property
    def src(self) -> EntityId:
        """Message source ID."""
        return self.header.src

    @property
    def dest(self) -> EntityId:
        """Message destination ID."""
        return self.header.dest

    @property
    def tag(self) -> uuid.UUID:
        """Message tag."""
        return self.header.tag

    @property
    def label(self) -> uuid.UUID | None:
        """Message label."""
        return self.header.label

    @classmethod
    def create(
        cls,
        src: EntityId,
        dest: EntityId,
        body: BodyT,
        *,
        label: uuid.UUID | None = None,
    ) -> Message[BodyT]:
        """Create a new message with the specified header and body.

        Args:
            src: Source entity ID.
            dest: Destination entity ID.
            body: Message body.
            label: Optional label for disambiguation.

        Returns:
            A new message instance.
        """
        if isinstance(body, get_args(Request)):
            kind = 'request'
        elif isinstance(body, get_args(Response)):
            kind = 'response'
        else:
            raise AssertionError('Unreachable.')
        header = Header(src=src, dest=dest, label=label, kind=kind)
        request: Message[BodyT] = Message(header=header, body=body)
        return request

    def create_response(self, body: ResponseT) -> Message[ResponseT]:
        """Create a response message from this request message.

        Args:
            body: Response message body.

        Returns:
            A new response message instance.

        Raises:
            ValueError: If this message is already a response.
        """
        header = self.header.create_response_header()
        response: Message[ResponseT] = Message(header=header, body=body)
        return response

    def get_body(self) -> BodyT:
        """Return the message body, deserializing if needed.

        Lazily deserializes and returns the body object.
        Caches the body to avoid redundant decoding.

        Returns:
            The deserialized body.
        """
        if isinstance(self.body, get_args(Body)):
            return self.body

        adapter: TypeAdapter[BodyT] = TypeAdapter(Body)
        body = (
            adapter.validate_json(self.body)
            if isinstance(self.body, str)
            else adapter.validate_python(self.body)
        )
        self.body = body
        return self.body

    def is_request(self) -> bool:
        """Check if the message is a request."""
        return self.header.kind == 'request'

    def is_response(self) -> bool:
        """Check if the message is a response."""
        return self.header.kind == 'response'

    @classmethod
    def model_deserialize(cls, data: bytes) -> Message[BodyT]:
        """Deserialize a message from bytes using pickle.

        Warning:
            This uses pickle and is therefore susceptible to all the
            typical pickle warnings about code injection.

        Args:
            data: The serialized message as bytes.

        Returns:
            The deserialized message instance.

        Raises:
            TypeError: If the deserialized object is not a Message.
        """
        message = pickle.loads(data)
        if not isinstance(message, cls):
            raise TypeError(
                'Deserialized message is not of type Message.',
            )
        return message

    def model_serialize(self) -> bytes:
        """Serialize the message to bytes using pickle.

        Warning:
            This uses pickle and is therefore susceptible to all the
            typical pickle warnings about code injection.

        Returns:
            The serialized message as bytes.
        """
        return pickle.dumps(self)



from __future__ import annotations

import asyncio
import fnmatch
from collections import defaultdict
from collections.abc import AsyncGenerator
from collections.abc import Generator
from typing import Any
from unittest import mock

import pytest


class MockRedis:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.values: dict[str, str] = {}
        self.lists: dict[str, list[str]] = defaultdict(list)
        self.events: dict[str, asyncio.Event] = defaultdict(asyncio.Event)

    async def aclose(self) -> None:
        pass

    async def blpop(
        self,
        keys: list[str],
        timeout: float = 0,
    ) -> list[str] | None:
        result: list[str] = []
        for key in keys:
            if len(self.lists[key]) > 0:
                item = self.lists[key].pop()
                self.events[key].clear()
            else:
                try:
                    await asyncio.wait_for(
                        self.events[key].wait(),
                        timeout=None if timeout == 0 else timeout,
                    )
                except asyncio.TimeoutError:
                    return None
                else:
                    item = self.lists[key].pop()
                    self.events[key].clear()
            result.extend([key, item])
        return result

    async def delete(self, key: str) -> None:  # pragma: no cover
        if key in self.values:
            del self.values[key]
        elif key in self.lists:
            self.lists[key].clear()

    async def exists(self, key: str) -> bool:  # pragma: no cover
        return key in self.values or key in self.lists

    async def get(
        self,
        key: str,
    ) -> str | list[str] | None:
        if key in self.values:
            return self.values[key]
        elif key in self.lists:
            raise NotImplementedError()
        return None

    async def lrange(self, key: str, start: int, end: int) -> list[str]:
        items = self.lists.get(key, None)
        if items is None:
            return []
        return items[start:end]

    async def ping(self, **kwargs) -> None:
        pass

    async def rpush(self, key: str, *values: str) -> None:
        for value in values:
            self.lists[key].append(value)
            self.events[key].set()

    async def scan_iter(self, pattern: str) -> AsyncGenerator[str]:
        for key in self.values:
            if fnmatch.fnmatch(key, pattern):
                yield key

    async def set(self, key: str, value: str) -> None:
        self.values[key] = value


@pytest.fixture
def mock_redis() -> Generator[None]:
    redis = MockRedis()
    with mock.patch('redis.asyncio.Redis', return_value=redis):
        yield



"""Exception types raised by exchange clients and servers."""

from __future__ import annotations


class ExchangeServerError(Exception):
    """Base exception type for exceptions raised by relay clients."""

    pass


class ForbiddenError(ExchangeServerError):
    """Client does not have correct permissions after authentication."""

    pass


class UnauthorizedError(ExchangeServerError):
    """Client is missing authentication tokens."""



"""Generate the code reference pages and navigation."""

from __future__ import annotations

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

PACKAGE_DIR = 'academy/'

for path in sorted(Path(PACKAGE_DIR).rglob('**/*.py')):
    module_path = path.with_suffix('')
    doc_path = path.relative_to(PACKAGE_DIR).with_suffix('.md')
    full_doc_path = Path('api', doc_path)

    parts = tuple(module_path.parts)
    parts = tuple('.'.join(parts[: i + 1]) for i in range(len(parts)))

    if parts[-1].endswith('__init__'):
        parts = parts[:-1]
        doc_path = doc_path.with_name('index.md')
        full_doc_path = full_doc_path.with_name('index.md')
    elif parts[-1].endswith('__main__'):
        continue

    mod_name = parts[-1].split('.')[-1]
    if mod_name.startswith('_'):
        continue

    nav_parts: tuple[str, ...]
    if len(parts) == 1:
        nav_parts = parts
    elif len(parts) == 2:  # noqa: PLR2004
        nav_parts = (parts[1],)
    else:
        nav_parts = tuple([parts[1]] + [p.split('.')[-1] for p in parts[2:]])
    nav[nav_parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, 'w') as fd:
        fd.write(f'# {parts[-1]}\n\n::: {parts[-1]}')

    mkdocs_gen_files.set_edit_path(full_doc_path, path)

with mkdocs_gen_files.open('api/SUMMARY.md', 'w') as nav_file:
    nav_file.writelines(nav.build_literate_nav())



from __future__ import annotations

import asyncio
import functools
import logging
import sys
import time
import uuid
from contextvars import ContextVar
from pickle import PicklingError
from typing import Any
from typing import Generic
from typing import TYPE_CHECKING
from typing import TypeVar
from weakref import WeakSet

if sys.version_info >= (3, 10):  # pragma: >=3.10 cover
    from typing import ParamSpec
else:  # pragma: <3.10 cover
    from typing_extensions import ParamSpec

from academy.exception import AgentTerminatedError
from academy.exception import ExchangeClientNotFoundError
from academy.identifier import AgentId
from academy.message import ActionRequest
from academy.message import ActionResponse
from academy.message import ErrorResponse
from academy.message import Message
from academy.message import PingRequest
from academy.message import ResponseT
from academy.message import ShutdownRequest
from academy.message import SuccessResponse

if TYPE_CHECKING:
    from academy.agent import AgentT
    from academy.exchange import ExchangeClient
else:
    # Agent is only used in the bounding of the AgentT TypeVar.
    AgentT = TypeVar('AgentT')

logger = logging.getLogger(__name__)

K = TypeVar('K')
P = ParamSpec('P')
R = TypeVar('R')

exchange_context: ContextVar[ExchangeClient[Any]] = ContextVar(
    'exchange_context',
)


class Handle(Generic[AgentT]):
    """Handle to a remote agent.

    Internally, handles use an
    [`ExchangeClient`][academy.exchange.ExchangeClient] to send requests to
    and receive responses from the remote agent. By default the correct
    exchange client is inferred from the context using a
    [context variable][contextvars] (specifically, the
    `academy.handle.exchange_context` variable). This allows the same handle
    to be used in different contexts, automatically using the correct client
    to send messages.

    When a handle is used in contexts that have not configured the exchange
    client (such as outside of an agent runtime or
    [`Manager`][academy.manager.Manager]), a default exchange can be provided
    via the `exchange` argument. For advanced usage, the `ignore_context` flag
    will cause the handle to only use the `exchange` argument no matter what
    the current context is.

    Note:
        The `exchange` argument will not be included when a handle is pickled.
        Thus, unpickled handles must be used in a context that configures
        an exchange client.

    Args:
        agent_id: ID of the remote agent.
        exchange: A default exchange client to be used if an exchange client
            is not configured in the current context.
        ignore_context: Ignore the current context and force use of `exchange`
            for communication.

    Raises:
        ValueError: If `ignore_context=True` but `exchange` is not provided.
    """

    def __init__(
        self,
        agent_id: AgentId[AgentT],
        *,
        exchange: ExchangeClient[Any] | None = None,
        ignore_context: bool = False,
    ) -> None:
        self.agent_id = agent_id
        self._exchange = exchange
        self._registered_exchanges: WeakSet[ExchangeClient[Any]] = WeakSet()
        self.ignore_context = ignore_context

        if ignore_context and not exchange:
            raise ValueError(
                'Cannot initialize handle with ignore_context=True '
                'and no explicit exchange.',
            )

        # Unique identifier for each handle object; used to disambiguate
        # messages when multiple handles are bound to the same mailbox.
        self.handle_id = uuid.uuid4()
        self._pending_response_futures: dict[
            uuid.UUID,
            asyncio.Future[Any],
        ] = {}
        self._shutdown_requests: set[uuid.UUID] = set()

        if self._exchange is not None:
            self._register_with_exchange(self._exchange)

    @property
    def exchange(self) -> ExchangeClient[Any]:
        """Exchange client used to send messages.

        Returns:
            Exchange client.

        Raises:
            ExchangeClientNotFoundError: If no exchange client is set in the
                current context nor was one provided to the handle.
        """
        if self.ignore_context:
            assert self._exchange is not None
            return self._exchange

        try:
            return exchange_context.get()
        except LookupError as e:
            if self._exchange is not None:
                return self._exchange

            raise ExchangeClientNotFoundError(self.agent_id) from e

    def __reduce__(
        self,
    ) -> tuple[
        type[Handle[Any]],
        tuple[Any, ...],
    ]:
        if self.ignore_context:
            raise PicklingError(
                'Handle with ignore_context=True is not pickle-able',
            )
        return (Handle, (self.agent_id,))

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}(agent_id={self.agent_id!r}, '
            f'exchange={self._exchange!r}, '
            f'ignore_context={self.ignore_context!r})'
        )

    def __str__(self) -> str:
        name = type(self).__name__
        return f'{name}<agent: {self.agent_id}>'

    def __getattr__(self, name: str) -> Any:
        async def remote_method_call(*args: Any, **kwargs: Any) -> R:
            return await self.action(name, *args, **kwargs)

        return remote_method_call

    async def _process_response(self, response: Message[ResponseT]) -> None:
        # Check if this is an error response for a shutdown request as those
        # are handled differently than other requests types (action, ping)
        # which always expect a response.
        if response.tag in self._shutdown_requests:
            self._shutdown_requests.remove(response.tag)
            body = response.get_body()
            assert isinstance(body, ErrorResponse)
            exception = body.get_exception()
            # The only ok error to be ignored is if the agent we intended to
            # shutdown was already shutdown.
            if (
                not isinstance(exception, AgentTerminatedError)
                or exception.uid != self.agent_id
            ):
                logger.error(
                    'Failure requesting shutdown for %s: %s (type: %s)',
                    self.agent_id,
                    exception,
                    type(exception),
                )
            return

        future = self._pending_response_futures.pop(response.tag)
        assert not future.cancelled()

        body = response.get_body()
        if isinstance(body, ActionResponse):
            future.set_result(body.get_result())
        elif isinstance(body, ErrorResponse):
            future.set_exception(body.exception)
        elif isinstance(body, SuccessResponse):
            future.set_result(None)
        else:
            raise AssertionError('Unreachable.')

    def _register_with_exchange(self, exchange: ExchangeClient[Any]) -> None:
        """Register to receive messages from exchange.

        Typically this will be called internally when sending a message.

        Args:
            exchange: Exchange client to listen to.
        """
        if exchange not in self._registered_exchanges:
            exchange.register_handle(self)
            self._registered_exchanges.add(exchange)

    async def action(self, action: str, /, *args: Any, **kwargs: Any) -> R:
        """Invoke an action on the agent.

        Args:
            action: Action to invoke.
            args: Positional arguments for the action.
            kwargs: Keywords arguments for the action.

        Returns:
            Result of the action.

        Raises:
            AgentTerminatedError: If the agent's mailbox was closed. This
                typically indicates the agent shutdown for another reason
                (it self terminated or via another handle).
            Exception: Any exception raised by the action.
        """
        exchange = self.exchange
        self._register_with_exchange(exchange)

        request = Message.create(
            src=exchange.client_id,
            dest=self.agent_id,
            label=self.handle_id,
            body=ActionRequest(action=action, pargs=args, kargs=kwargs),
        )
        loop = asyncio.get_running_loop()
        future: asyncio.Future[R] = loop.create_future()
        self._pending_response_futures[request.tag] = future

        await self.exchange.send(request)
        logger.debug(
            'Sent action request from %s to %s (action=%r)',
            self.client_id,
            self.agent_id,
            action,
        )
        await future
        return future.result()

    async def ping(self, *, timeout: float | None = None) -> float:
        """Ping the agent.

        Ping the agent and wait to get a response.

        Args:
            timeout: Optional timeout in seconds to wait for the response.

        Returns:
            Round-trip time in seconds.

        Raises:
            AgentTerminatedError: If the agent's mailbox was closed. This
                typically indicates the agent shutdown for another reason
                (it self terminated or via another handle).
            TimeoutError: If the timeout is exceeded.
        """
        exchange = self.exchange
        self._register_with_exchange(exchange)

        request = Message.create(
            src=exchange.client_id,
            dest=self.agent_id,
            label=self.handle_id,
            body=PingRequest(),
        )
        loop = asyncio.get_running_loop()
        future: asyncio.Future[None] = loop.create_future()
        self._pending_response_futures[request.tag] = future
        start = time.perf_counter()
        await self.exchange.send(request)
        logger.debug('Sent ping from %s to %s', self.client_id, self.agent_id)

        done, pending = await asyncio.wait({future}, timeout=timeout)
        if future in pending:
            raise TimeoutError(
                f'Did not receive ping response within {timeout} seconds.',
            )
        elapsed = time.perf_counter() - start
        logger.debug(
            'Received ping from %s to %s in %.1f ms',
            exchange.client_id,
            self.agent_id,
            elapsed * 1000,
        )
        return elapsed

    async def shutdown(self, *, terminate: bool | None = None) -> None:
        """Instruct the agent to shutdown.

        This is non-blocking and will only send the request. Any unexpected
        error responses sent by the exchange will be logged.

        Args:
            terminate: Override the termination behavior of the agent defined
                in the [`RuntimeConfig`][academy.runtime.RuntimeConfig].

        Raises:
            AgentTerminatedError: If the agent's mailbox was closed. This
                typically indicates the agent shutdown for another reason
                (it self terminated or via another handle).
        """
        exchange = self.exchange
        self._register_with_exchange(exchange)

        request = Message.create(
            src=exchange.client_id,
            dest=self.agent_id,
            label=self.handle_id,
            body=ShutdownRequest(terminate=terminate),
        )
        self._shutdown_requests.add(request.tag)
        await self.exchange.send(request)
        logger.debug(
            'Sent shutdown request from %s to %s',
            exchange.client_id,
            self.agent_id,
        )


class ProxyHandle(Handle[AgentT]):
    """Proxy handle.

    A proxy handle is thin wrapper around an
    [`Agent`][academy.agent.Agent] instance that is useful for testing
    agents that are initialized with a handle to another agent without
    needing to spawn agents. This wrapper invokes actions synchronously.
    """

    def __init__(self, agent: AgentT) -> None:
        self.agent = agent
        self.agent_id: AgentId[AgentT] = AgentId.new()
        self._agent_closed = False

    def __repr__(self) -> str:
        return f'{type(self).__name__}(agent={self.agent!r})'

    def __str__(self) -> str:
        return f'{type(self).__name__}<{self.agent}>'

    def __getattr__(self, name: str) -> Any:
        method = getattr(self.agent, name)
        if not callable(method):
            raise AttributeError(
                f'Attribute {name} of {type(self.agent)} is not a method.',
            )

        @functools.wraps(method)
        async def func(*args: Any, **kwargs: Any) -> R:
            return await self.action(name, *args, **kwargs)

        return func

    def __reduce__(
        self,
    ) -> tuple[
        type[Handle[Any]],
        tuple[Any, ...],
    ]:
        return (ProxyHandle, (self.agent,))

    async def action(self, action: str, /, *args: Any, **kwargs: Any) -> R:
        """Invoke an action on the agent.

        Args:
            action: Action to invoke.
            args: Positional arguments for the action.
            kwargs: Keywords arguments for the action.

        Returns:
            Result of the action.

        Raises:
            AgentTerminatedError: If the agent's mailbox was closed. This
                typically indicates the agent shutdown for another reason
                (it self terminated or via another handle).
            Exception: Any exception raised by the action.
        """
        if self._agent_closed:
            raise AgentTerminatedError(self.agent_id)

        method = getattr(self.agent, action)
        return await method(*args, **kwargs)

    async def ping(self, *, timeout: float | None = None) -> float:
        """Ping the agent.

        This is a no-op for proxy handles and returns 0 latency.

        Args:
            timeout: Optional timeout in seconds to wait for the response.

        Returns:
            Round-trip time in seconds.

        Raises:
            AgentTerminatedError: If the agent's mailbox was closed. This
                typically indicates the agent shutdown for another reason
                (it self terminated or via another handle).
            TimeoutError: If the timeout is exceeded.
        """
        if self._agent_closed:
            raise AgentTerminatedError(self.agent_id)
        return 0

    async def shutdown(self, *, terminate: bool | None = None) -> None:
        """Instruct the agent to shutdown.

        This is non-blocking and will only send the message.

        Args:
            terminate: Override the termination behavior of the agent defined
                in the [`RuntimeConfig`][academy.runtime.RuntimeConfig].

        Raises:
            AgentTerminatedError: If the agent's mailbox was closed. This
                typically indicates the agent shutdown for another reason
                (it self terminated or via another handle).
        """
        if self._agent_closed:
            raise AgentTerminatedError(self.agent_id)
        self._agent_closed = True if terminate is None else terminate
